
/**
 * Title:        Approximations Simulation<p>
 * Description:  Simulation sur des exemples de divers algorithmes d'approximation pour la compression de chanes de caractres utilisant des grammaires non contextuelles.<p>
 * Copyright:    Copyright (c) Ccile Malinaud/Yann Ponty<p>
 * Company:      <p>
 * @author Ccile Malinaud/Yann Ponty
 * @version 1.0
 */

 /**
  * Les grammaires instances de cette classe forment un sous-ensemble strict des grammaires
  * non contextuelles. A partir d'une grammaire, on ne peut gnrer qu'une seule
  * squences de symboles terminaux. La taille d'une telle grammaire correspond
  *  un codage binaire de cette grammaire permettant d'engendrer son langage.
  * Ex. :
  *  G :
  *   T -> a R
  *   R -> b b
  *  On va attribuer des ensembles d'indices disjoints aux terminaux et non-terminaux :
  *   I(a) = 0
  *   I(b) = 1
  *   I(T) = 2
  *   I(R) = 3
  *  On se munit d'un caractre spcial '#' permettant de sparer les rgles de la grammaire.
  *  De plus, on remarquera que, la grammaire tant non contextuelle, les parties droites de ses
  *  rgles sont rduites  un seul symbole, on n'aura donc pas  sparer les parties droites et
  *  gauches des rgles.
  *  On en arrive au codage suivant :
  *  C(G) :
  *   203#311
  *  Un tel codage conduit  une valuation de la taille d'une grammaire :
  *    taille(G) = #occurencesSymbolesDansPartieDroiteDeRegle*log2(#symboles)
  */
package approximations;

import java.util.*;

public class Grammar {

  private Vector _rules = new Vector();

  private long _axiom = 0;

  public Grammar() {
  }

  /**
   * Ajoute une rgle  la grammaire.
   * @params partieGauche un entier < 0 codant le symbole non terminal
   * @params partieDroite un tableau d'entiers dont les contenus ngatifs
   * reprsentent des symboles non terminaux et les positifs des caractres
   * (symboles terminaux)
   */
  public void addRule(long partieGauche, long[] partieDroite)
  {
    if(partieGauche<0)
    {
      Couple c = new Couple(new Long(partieGauche),partieDroite);
      int index = Collections.binarySearch(_rules,c);
      if(index<0)
      {
        _rules.add((-index)-1,c);
        if (_axiom==0){_axiom = partieGauche;}
      }
    }
  }

  public boolean containsRule(long partieGauche)
  {
    if(partieGauche<0)
    {
      Couple c = new Couple(new Long(partieGauche),null);
      int index = Collections.binarySearch(_rules,c);
      if(index>=0)
      {
        return true;
      }
    }
    return false;
  }


  public long[] getRule(long partieGauche)
  {
    if(partieGauche<0)
    {
      Couple c = new Couple(new Long(partieGauche),null);
      int index = Collections.binarySearch(_rules,c);
      if(index>=0)
      {
        return (long[]) (((Couple)_rules.elementAt(index)).snd());
      }
    }
    return new long[0];
  }

  public void alterRule(long partieGauche, long[] partieDroite)
  {
    if(partieGauche<0)
    {
      Couple c = new Couple(new Long(partieGauche),partieDroite);
      int index = Collections.binarySearch(_rules,c);
      if(index>=0)
      {
        ((Couple)_rules.elementAt(index)).setSnd(partieDroite);
      }
    }
  }

  public void setAxiom(long axiom)
  {
    int index = Collections.binarySearch(_rules,new Couple(new Long(axiom),null));
    if(index>=0)
    {
      _axiom = axiom;
    }
  }

  public String getLanguageFor(long nt)
  {
    String result = "";
    if(nt<0)
    {
      Couple c = new Couple(new Long(nt),null);
      int index = Collections.binarySearch(_rules,c);
      if(index>=0)
      {
        Vector sequence = new Vector();
        sequence.add(new Long(nt));
        int i = 0;
        while(i<sequence.size())
        {
          Long l = (Long) sequence.elementAt(i);
          if (l.longValue()<0)
          {
            sequence.remove(i);
            index = Collections.binarySearch(_rules,new Couple(l,null));
            if (index>=0)
            {
              long[] RHS = (long[])((Couple) _rules.elementAt(index)).snd();
              for(int j=RHS.length-1;j>=0;j--)
              { sequence.add(i,new Long(RHS[j])); }
            }
            else
            {
              result += "\nNonTerminal Symbol X"+(-l.intValue())+" is not defined !\n";
            }
          }
          else
          {
            result+=""+(char)l.intValue();
            i++;
          }
        }
      }
    }
    return result;
  }

  public String getLanguage()
  {
    String result = "";
    Vector sequence = new Vector();
    sequence.add(new Long(_axiom));
    int i = 0;
    while(i<sequence.size())
    {
      Long l = (Long) sequence.elementAt(i);
      if (l.longValue()<0)
      {
        sequence.remove(i);
        int index = Collections.binarySearch(_rules,new Couple(l,null));
        if (index>=0)
        {
          long[] RHS = (long[])((Couple) _rules.elementAt(index)).snd();
          for(int j=RHS.length-1;j>=0;j--)
          { sequence.add(i,new Long(RHS[j])); }
        }
        else
        {
          result += "[Error : NT Symbol X"+(-l.intValue())+" is not defined]";
        }
      }
      else
      {
        result+=""+(char)l.intValue();
        i++;
      }
    }
    return result;
  }

  public String toString()
  {
    return toString(false);
  }

  public String toString(boolean printRules)
  {
    String result = "";
    int size=0;
    for(int i=0;i<_rules.size();i++)
    {
      Couple c = (Couple) _rules.elementAt(i);
      long LHS = ((Long)c.fst()).longValue();
      long[] RHS = (long[])c.snd();
      size += RHS.length;
      if (printRules)
      {
        result += "X"+(-LHS)+" -> ";
        for(int j=0;j<RHS.length;j++)
        {

          if(RHS[j]>=0)
          {
            // Si le caractre est affichable, on l'affiche
            if(RHS[j]>=20)
            result += (char)RHS[j]+" ";
            else
            result += "#"+(int)RHS[j]+" ";
          }
          else
          {
            result += "X"+(-RHS[j])+" ";
          }
        }
        result +="\n";
      }
    }
    result +="  Taille : "+size;
    return result;
  }

}